#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Copyright (C) 2006  Fabian Steiner <info@fabis-site.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import os
import sys
import select
import time
import cPickle
import termios
from string import Template
from optparse import OptionParser
from multiseat.utils import *
from multiseat.output import Output
from multiseat.seat import Seat, TemplateParserException
from multiseat.constants import *

############################################################
# Check whether we are root or not
############################################################

if os.getuid() != 0:
    print >> sys.stderr, 'You have to run this script as root. Your UID is %s' % os.getuid()
    sys.exit(1)

############################################################
# variable definitions
############################################################

settings = {}
collection = []

############################################################
# command line arguments proecessing
############################################################

parser = OptionParser(version='%prog ' + VERSION)
parser.add_option('-c', '--nocolour', dest='colour', action='store_false', help='do not use colours to format ouput')
parser.add_option('-d', '--debug', dest='debug', action='store_true', help='enable debug mode')
parser.add_option('-p', '--nopickling', dest='pickle', action='store_false', help='do not pickle current input device configuration')
parser.add_option('-r', '--read', dest='read', action='store_true', help='read the pickled file and re-create xorg.conf')
parser.add_option('-s', '--seats', dest='seats', type='int', help='configure input devices for the given number of displays')
parser.add_option('-w', '--wait', dest='wait', type='float', help='wait the given number of seconds for input')

parser.set_defaults(colour=True,       # we want coloured output
                              debug=False,      # be verbose
                              pickle=True,        # save the configuration of input devices in a pickled file
                              read=False,         # don't read the pickled file per default
                              wait=5.0)            # you have got 5.0 seconds to press a key / mouve the mouse
(options, args) = parser.parse_args()

if not options.read and not options.seats:
    parser.error('You have to define the number of seats.')

settings['colour'] = options.colour
settings['debug'] = options.debug
settings['pickle'] = options.pickle
settings['read_pickled_file'] = options.read
settings['seats'] = options.seats
settings['timeout'] = options.wait
    
############################################################
# create instance of the Output class
# (used to format the output to define whether to be verbose or not)
############################################################

output = Output(debug=settings['debug'], use_colour=settings['colour'])

############################################################
# we simply read the pickled file and don't poll the input devices
############################################################

if settings['read_pickled_file']:
    output.dbg('Reading pickled file to restore input device configuration')
    try:
        f = file(PICKLED_FILE, 'rb')
    except IOError, errstr:
        output.err('Can not open %s for reading. Error: %s' % (PICKLED_FILE, errstr))
        sys.exit(1)
    collection = cPickle.load(f)
    f.close()
else:
    ##########################################################
    # polling of input devices
    ##########################################################  
    
    # get a collection of available video cards by parsing the output of `lspci`
    video_cards = get_video_cards()
    
    if not video_cards:
        output.err('There are no video cards available. Aborting')
        sys.exit(1)
    elif len(video_cards) < settings['seats']:
        output.err('There is/are only %d video card(s) available, but you specified %d seats.' %  
            (len(video_cards), settings['seats']))
        sys.exit(1)
    
    output.dbg('Configuring input devices for %d displays' % settings['seats'])
      
    # now we will walk through the number of specified seats and probe for keyboard and mouse    
    for seat in range(settings['seats']):
        seat_obj = Seat()
        for attr, attr_value in {'keyboard_identifier': 'Keyboard', 
                                            'mouse_identifier': 'Mouse', 
                                            'video_card_identifier': 'VideoCard',
                                            'monitor_identifier': 'Monitor',
                                            'screen_identifier': 'Screen',
                                            'layout_identifier': 'Layout'
                                            }.iteritems():
            setattr(seat_obj, attr, attr_value + str(seat))
        seat_obj.screen_number = '0'
        
        print 'Which video card do you want to associate with display :%d.0?' % seat
        for counter, card in enumerate(video_cards):
            print '[%d] %s' % (counter, card)
        while 1:
            stdin_fd = sys.stdin.fileno()
            termios.tcflush(stdin_fd, termios.TCIFLUSH)
            choice = raw_input('Your choice: ').strip()
            try:
                selected_card = video_cards[int(choice)]
            except ValueError:
                output.err('Can not convert your choice to int.')
                continue
            except IndexError:
                output.err('Your choice is not available.')
                continue
            seat_obj.video_card_busid = selected_card
            del stdin_fd, video_cards[int(choice)], choice
            break
        
        for input_device in  ('keyboard', 'mouse'):
            if input_device == 'keyboard':
                print
                print 'Now we are about to configure the keyboard for display :%s.0' % seat
                print 'Please press some _keys_ on the keyboard which should be associated with display :%s.0' % seat
                # we wait some time so that everbody should be able to read the instructions above
                output.countdown(5)
                print
            elif input_device == 'mouse':
                print
                print 'Now we are about to configure the mouse for display :%s.0' % seat
                print 'Please move the _mouse_ which should be associated with display :%s.0' % seat
                # we wait some time so that everbody should be able to read the instructions above
                output.countdown(5)
                print
                
            # this dictionary will hold all opened file descriptors
            watchfds = {}
            
            for device in get_all_input_devices():
                try:
                    # open the evdev input devices (/dev/input/event0) readonly to get_all_input_devices
                    # the file descriptors which we will pass to the select() function
                    fd = os.open(device['handler'], os.O_RDONLY)
                except OSError, errstr:
                    output.err('%s: %s' % (device['handler'], errstr))
                    sys.exit(1)
                
                watchfds[fd] = ({'handler': device['handler'],      # evdev event interface (e.g. /dev/input/event0)
                                                'phys': device['phys'],                   # phys path of the input device
                                                'activity': 0                                      # activity (will be incremented by 2  --> key press,
                                            })                                                                 # key release)
            
            # start time (this is needed to find out how long we should wait for
            # events of the input devices)
            start = time.time()
            
            while 1:
                rfd, wfd, err = select.select(watchfds.keys(), [], [], settings['timeout'])
                
                if rfd:
                    for fd in rfd:
                        if fd in watchfds:
                            watchfds[fd]['activity'] += 1
                            output.dbg('%s - %d' % (watchfds[fd]['handler'], 
                                                               watchfds[fd]['activity']))
                            os.read(fd, 10000)
            
                stop = time.time()
                
                if stop >= (start + settings['timeout']):
                    # obviously, we have been waiting for <settings['timeout']> seconds, we leave
                    # the loop now in order to continue with the other devices
                    break
            
            max = {'handler': '', 'phys':  '', 'activity': 0}
                    
            for fd in watchfds:
                if watchfds[fd]['activity'] > 0:
                    if watchfds[fd]['activity'] > max['activity']:
                        # the current device has got a higher activity so it is our new
                        # 'max-rated' input device
                        max = watchfds[fd] 
                    print 'final count: %s - %s' % (watchfds[fd]['handler'], 
                                                                                       watchfds[fd]['activity'])
                os.close(fd)
            
            if max['activity'] > 0:
                print '%s is current best candidate: %d' % (max['handler'], 
                                                                                           max['activity'])
                
                if input_device == 'keyboard':
                    seat_obj.keyboard_device = max['handler']
                    seat_obj.keyboard_phys = max['phys']
                elif input_device == 'mouse':
                   seat_obj.mouse_device = max['handler'] 
                   seat_obj.mouse_phys = max['phys']
            else:
                output.err("Timeout")
                sys.exit(1)
            
            del watchfds

        collection.append(seat_obj)
    
    print 'Probing input devices succesfully finished.'
    
    ############################################################
    # pickling the configuration of the input devices if the user wants us to do it
    ############################################################
    
    if settings['pickle']:
        try:
            f = file(PICKLED_FILE, 'wb')
        except IOError, errstr:
            output.err('Can not write configuration to %s. Error: %s' % (PICKLED_FILE, errstr))
            sys.exit(1)  
        cPickle.dump(collection, f)
        f.close()
        output.dbg('Pickled configuration successfully written to %s' % PICKLED_FILE)

############################################################
# generating proper xorg.conf by using an appropriate template file
############################################################

print 'Editing the xorg.conf template file'

try:
    f = file(STATIC_TEMPLATE)
except IOError, errstr:
    output.err('Can not open %s for reading. Error: %s' % (STATIC_TEMPLATE, errstr))
    sys.exit(1)
static_template = f.read()
f.close()

seat_template = ''
for counter, seat_obj in enumerate(collection):
    try:
        seat_template += '### WORKSTATION %d ###\n\n' % counter
        seat_template += seat_obj.substitute()
        seat_template += '\n'
    except TemplateParserException, err:
        output.err(err.msg)
        sys.exit(1)

template = Template(static_template).substitute(SEATS=seat_template)

f = file(XORG_CONFIG, 'w')
f.write(template)
f.close()

output.dbg('Modified Template successfully written to %s' % XORG_CONFIG)

sys.exit(0)
